unit Pipe;

{
Author:       Juergen Ruemmler
EMail:        medicus@cityweb.de
Version:      1.1. Stamd 26.02.1999
Legal issues: Copyright (C) 1998, 1999 by Juergen Ruemmler
              Brockdorffstr. 27a, 22149 Hamburg, Deutschland

              TPipe ist Freeware.

Beschreibung: TPipe dient als Zwischenspeicher nach dem FiFo-Prinzip:
              Daten, die als erstes eingespeichert wurden, kommen
              auch als erstes wieder raus.
              Dabei wird Datentyp und -Laenge nicht bercksichtigt.

  Daten speichern:  Neue Daten werden ans Ende angefuegt.

  Daten lesen:      Daten werden vom Anfang der Pipe
              --->  gelesen. Gelesene Daten werden geloescht
              --->  und verbleibende Daten werden an den
              --->  Anfang verschoben, und das ist der
              --->  gravierende Unterschied zu TMemoryStream !!!

                    1. Ausnahme: GetByte(Position)
                    Position = 0..(Size-1)
                    Die Funktion dient dazu testweise ein Byte
                    zu lesen ohne den Inhalt der Pipe zu aendern.
                    2. Ausnahme: CopyToBuffer
                    Daten werden in Buffer geschrieben aber in Pipe
                    nicht geloescht.

                    Rueckgabewert der Functionen WriteTo..
                    ist die tatsaechliche Anzahl Daten, die
                    zurueckgegeben werden.

Historie:

  Version    Datum        Kommentar
  -------    ---------    ---------
  1.0        16.06.1998   Basis-Version
  1.1        26.02.1999   WriteToPipe hinzugefgt
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dsgnintf, Dialogs;

type
  TPipe = class(TComponent)
  private
    DataAnz: Integer;
    Data: Pointer;
    FAbout: String;
    function peek(Adresse: Integer): Byte; pascal;
    procedure poke(Adresse: Integer; Wert: Byte); pascal;
    function Adresse(x: Pointer): Integer;
    procedure CopyMem(Source, Dest, Count: Integer);
  public
    // Alle Funktionen geben weder Buffer-Speicher frei
    // noch reservieren sie welchen !!!
    property PipeSize: Integer Read DataAnz;
    procedure Clear;
    procedure ShowAbout;
    procedure ReadFromBuffer(Buff: Pointer; Size: Integer);
    procedure ReadFromStream(Stream: TStream; Size: Integer);
    procedure ReadFromString(h: String);
    function GetByte(Position: Integer): Byte;
    function WriteToBuffer(Buffer: Pointer; Size: Integer): Integer;
    procedure CopyToBuffer(Buffer: Pointer; Size: Integer);
    function WriteToStream(Stream: TStream; Size: Integer): Integer;
    function WriteToString(var h: String; Size: Integer): Integer;
    function WriteToPipe(ToPipe: TPipe; Size: Integer): Integer;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property About: String read FAbout write FAbout stored False;
  end;

procedure Register;

implementation

{*.dcr}

type
    TAboutProperty = class(TPropertyEditor)
    public
      procedure Edit; override;
      function GetAttributes: TPropertyAttributes; override;
      function GetValue: String; override;
    end;

procedure Register;
begin
  RegisterComponents('System', [TPipe]);
  RegisterPropertyEditor(TypeInfo(String), TPipe, 'About', TAboutProperty);
end;

procedure TAboutProperty.Edit;
begin
  TPipe(GetComponent(0)).ShowAbout;
end;

function TAboutProperty.GetAttributes: TPropertyAttributes;
begin
  GetAttributes:=[paDialog, paReadOnly];
end;

function TAboutProperty.GetValue: String;
begin
  GetValue:='(About)';
end;

procedure TPipe.ShowAbout;
var msg: String;
const cr = #13;
begin
  msg:='TPipe v1.1' + cr;
  msg:=msg + 'Stand: 26.02.1999' + cr;
  msg:=msg + 'Eine Freeware-Komponente' + cr;
  msg:=msg + 'entwickelt in Delphi 3' + cr + cr;
  msg:=msg + 'Copyright  1998, 1999 Juergen Ruemmler, Germany' + cr;
  msg:=msg + 'eMail: medicus@cityweb.de';
  ShowMessage(msg);
end;

function TPipe.Adresse(x: Pointer): Integer;
begin
  asm
    MOV    EAX, x
    MOV    @Result, EAX
  end;
end;

function TPipe.Peek(Adresse: Integer): Byte; pascal;
begin
  asm
    MOV  EDX, Adresse
    MOV  AL, [EDX]
    MOV  @Result, AL
  end;
end;

procedure TPipe.Poke(Adresse: Integer; Wert: Byte); pascal;
begin
  asm
    MOV  AL, Wert
    MOV  EDX, Adresse
    MOV  [EDX], AL
  end;
end;

procedure TPipe.CopyMem(Source, Dest, Count: Integer);
begin
  if Source > Dest then   // aufwrts kopieren
    asm
      MOV   ECX, Source
      MOV   EDX, Dest
      CMP   Count, 0
      JE    @ende
    @schl1:
      MOV   AL, [ECX]
      MOV   [EDX], AL
      ADD   ECX, 1
      ADD   EDX, 1
      SUB   Count, 1
      CMP   Count, 0
      JNE   @schl1
    @ende:
    end
  else                    // abwrts kopieren
    asm
      MOV   ECX, Source
      MOV   EDX, Dest
      ADD   ECX, Count
      ADD   EDX, Count
      SUB   ECX, 1
      SUB   EDX, 1
      CMP   Count, 0
      JE    @ende
    @schl1:
      MOV   AL, [ECX]
      MOV   [EDX], AL
      SUB   ECX, 1
      SUB   EDX, 1
      SUB   Count, 1
      CMP   Count, 0
      JNE   @schl1
    @ende:
    end;
end;

constructor TPipe.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Data:=Nil;
  DataAnz:=0;
end;

destructor TPipe.Destroy;
begin
  if not (Data = Nil) then FreeMem(Data);
  inherited Destroy;
end;

procedure TPipe.Clear;
begin
  ReAllocMem(Data, 0);
  DataAnz:=0;
end;

procedure TPipe.ReadFromBuffer(Buff: Pointer; Size: Integer);
// Daten werden aus dem Buffer BUFF in die Pipe kopiert
begin
  try
    ReAllocMem(Data, DataAnz + Size);
  except
    raise Exception.Create('Not enough Memory!');
    exit;
  end;
  CopyMem(Adresse(Buff), Adresse(Data) + DataAnz, Size);
  DataAnz:=DataAnz + Size;
end;

procedure TPipe.ReadFromStream(Stream: TStream; Size: Integer);
// Daten werden aus dem Stream in die Pipe kopiert
// Der Kopiervorgang von der Stream-Komponente vorgenommen
var p: Pointer;
begin
  try
    ReAllocMem(Data, DataAnz + Size);
  except
    raise Exception.Create('Not enough Memory!');
    exit;
  end;
  p:=Ptr(Adresse(Data) + DataAnz);
  Stream.ReadBuffer(p, Size);
  DataAnz:=DataAnz + Size;
end;

procedure TPipe.ReadFromString(h: String);
// Daten werden Byte fr Byte aus dem String in die Pipe
// kopiert
var i,a: integer;
begin
  try
    ReAllocMem(Data, DataAnz + Length(h));
  except
    raise Exception.Create('Not enough Memory!');
    exit;
  end;
  a:=Adresse(Data);
  for i:=1 to Length(h) do
    Poke(a + DataAnz + i - 1, Ord(h[i]));
  DataAnz:=DataAnz + Length(h);
end;

function TPipe.GetByte(Position: Integer): Byte;
// Liefert ein wahlfreies Byte zurck, ohne es in der Pipe
// zu lschen !!!
// Position = 0..(PipeSize - 1)
begin
  if Position + 1 > DataAnz then Position := DataAnz - 1;
  Result:=Peek(Adresse(Data) + Position);
end;

function TPipe.WriteToBuffer(Buffer: Pointer; Size: Integer): Integer;
// Die ersten [Size]-Bytes werden in den Buffer kopiert und aus der
// Pipe gelscht. Die aufrufende Routine mu den Speicher fr BUFFER
// bereit stellen
begin
  if Size > DataAnz then Size:=DataAnz;
  if Size < 0 then begin
    WriteToBuffer:=0;
    exit;
  end;
  CopyMem(Adresse(Data), Adresse(Buffer), Size);
  DataAnz:=DataAnz - Size;
  CopyMem(Adresse(Data) + Size, Adresse(Data), DataAnz);
  ReAllocMem(Data, DataAnz);
  WriteToBuffer:=Size;
end;

procedure TPipe.CopyToBuffer(Buffer: Pointer; Size: Integer);
// Die ersten [Size]-Bytes werden in den Buffer kopiert.
begin
  if Size > DataAnz then Size:=DataAnz;
  if Size < 0 then begin
    exit;
  end;
  CopyMem(Adresse(Data), Adresse(Buffer), Size);
end;

function TPipe.WriteToStream(Stream: TStream; Size: Integer): Integer;
// Die ersten [Size]-Bytes werden in den Stream kopiert und aus der
// Pipe gelscht. Der Kopiervorgang wird von den Stream-Routinen
// vorgenommen.
var p: Pointer;
begin
  if Size > DataAnz then Size:=DataAnz;
  if Size < 0 then begin
    WriteToStream:=0;
    exit;
  end;
  p:=Ptr(Adresse(Data));
  Stream.WriteBuffer(p, Size);
  DataAnz:=DataAnz - Size;
  CopyMem(Adresse(Data) + Size, Adresse(Data), DataAnz);
  ReAllocMem(Data, DataAnz);
  WriteToStream:=Size;
end;

function TPipe.WriteToPipe(ToPipe: TPipe; Size: Integer): Integer;
begin
  if Size > DataAnz then Size:=DataAnz;
  if Size < 0 then begin
    WriteToPipe:=0;
    exit;
  end;
  // in ToPipe fr Platz sorgen ...
  try
    ReAllocMem(ToPipe.Data, ToPipe.DataAnz + Size);
  except
    raise Exception.Create('Not enough Memory!');
    exit;
  end;
  // von "dieser Pipe" nach ToPipe kopieren
  CopyMem(Adresse(Data), Adresse(ToPipe.Data) + ToPipe.DataAnz, Size);
  ToPipe.DataAnz:=ToPipe.DataAnz + Size;
  // in "dieser Pipe" gelesene Daten lschen
  DataAnz:=DataAnz - Size;
  CopyMem(Adresse(Data) + Size, Adresse(Data), DataAnz);
  ReAllocMem(Data, DataAnz);
  WriteToPipe:=Size;
end;

function TPipe.WriteToString(var h: String; Size: Integer): Integer;
// Die ersten [Size]-Bytes werden in den String kopiert und aus der
// Pipe gelscht. Der String wir vor Beginn gelscht !!!
var i,a: integer;
begin
  if Size > DataAnz then Size:=DataAnz;
  if Size < 0 then begin
    WriteToString:=0;
    exit;
  end;
  h:='';
  a:=Adresse(Data);
  for i:=1 to Size do
    h:=h + Chr(Peek(a + i - 1));
  DataAnz:=DataAnz - Size;
  WriteToString:=Size;
  CopyMem(Adresse(Data) + Size, Adresse(Data), DataAnz);
  ReAllocMem(Data, DataAnz);
end;

end.
